from typing import Optional, Type, Union

import numpy as np
import polars as pl
from typing_extensions import final


@final
class _Unset:
    pass


def pl_index(df: pl.DataFrame, indices: np.ndarray, index_col: str = "index") -> pl.DataFrame:
    assert index_col not in df.columns

    return (
        df.with_row_index(index_col)
        .join(
            pl.DataFrame({index_col: indices}),
            on=index_col,
            how="inner",
            validate="1:1",
            maintain_order="left",
        )
        .drop(index_col)
    )


def pl_islice(df: pl.DataFrame, start: Optional[int], stop: Union[int, None, Type[_Unset]] = _Unset) -> pl.DataFrame:
    if start is None and stop is None:
        return df
    elif stop is _Unset:
        return df.slice(0, start)
    elif start is None:
        return df.slice(0, stop)
    elif stop is None:
        return df.slice(start, None)
    else:
        return df.slice(start, stop - start)



import os
from multiprocessing import Manager, RLock
from multiprocessing.synchronize import RLock as RLockT
from threading import Lock
from typing import Any, Iterable, Iterator, MutableMapping, Optional, Sequence, TextIO, Tuple, Type, TypeVar, Union

from typing_extensions import Self

from tqdm import tqdm

from .callbacks import BaseTask
from .callbacks import Progress as _Progress
from .callbacks import _Default

T = TypeVar("T")


class Task(BaseTask):
    def __init__(
        self, total: Optional[float] = None, description: Optional[str] = None, transient: bool = False, **kwargs: Any
    ) -> None:
        leave = not transient
        self.pbar = tqdm(desc=description, total=total, leave=leave, **kwargs)

    def __enter__(self) -> Self:
        return self

    def __exit__(self, *args):
        self.pbar.close()

    def advance(self, delta: float) -> None:
        self.pbar.update(n=delta)

    def update(
        self,
        *,
        completed: Union[Optional[float], Type[_Default]] = _Default,
        total: Union[Optional[float], Type[_Default]] = _Default,
        description: Union[Optional[str], Type[_Default]] = _Default,
        **kwargs: Any,
    ) -> None:
        if completed is not _Default:
            self.pbar.n = completed
        if total is not _Default:
            self.pbar.total = total
        if description is not _Default:
            self.pbar.set_description(description)


class Progress(_Progress):
    def track(
        self,
        sequence: Union[Iterable[T], Sequence[T]],
        total: Optional[float] = None,
        description: Optional[str] = None,
        transient: bool = False,
        **kwargs: Any,
    ) -> Iterator[T]:
        leave = not transient
        yield from tqdm(sequence, description, total, leave, **kwargs)

    def task(
        self, total: Optional[float] = None, description: Optional[str] = None, transient: bool = False, **kwargs: Any
    ) -> BaseTask:
        return Task(total, description, transient, **kwargs)

    def set_prolog(self, prolog: str) -> None:
        raise NotImplementedError

    def set_epilog(self, epilog: str) -> None:
        raise NotImplementedError

    def print(self, s: str, end="\n", **kwargs: Any) -> None:
        tqdm.write(s, end=end, nolock=kwargs.get("nolock", False))


class TqdmProcess:
    def __init__(self, lock: Lock, pids: MutableMapping[int, int]) -> None:
        self.lock = lock
        self.pids = pids

    def track(
        self,
        iterable: Optional[Iterable] = None,
        desc: Optional[str] = None,
        total: Union[int, float, None] = None,
        transient: bool = False,
        file: Optional[TextIO] = None,
        ncols: Optional[int] = None,
        mininterval: float = 0.1,
    ):
        with self.lock:
            position = self.pids.setdefault(os.getpid(), len(self.pids))

        leave = not transient
        yield from tqdm(iterable, desc, total, leave, file, ncols, mininterval, position=position)

    def write(self, s: str, file=None, end="\n", nolock=False) -> None:
        tqdm.write(s, file, end, nolock)

    def initializer(self, lock: RLockT) -> None:
        tqdm.set_lock(lock)

    @property
    def initargs(self) -> Tuple[RLockT]:
        return (tqdm.get_lock(),)


class TqdmMultiprocessing:
    def __init__(self) -> None:
        self.manager = Manager()
        tqdm.set_lock(RLock())

    def __enter__(self) -> TqdmProcess:
        manager = self.manager.__enter__()
        lock = manager.Lock()
        pids = manager.dict()
        return TqdmProcess(lock, pids)

    def __exit__(self, *args):
        self.manager.__exit__(*args)



"""This module is to avoid circular imports.
It should avoid any dependencies apart from the standard library.
"""

import os
import os.path
import re
from typing import Optional, Union

from .typing import DirEntryProtocol

PathType = Union[str, os.PathLike]


class _DirEntry(os.PathLike):
    __slots__ = ("entry",)

    def __init__(
        self,
        name: str,
        path: Optional[str] = None,
        inode: Optional[int] = None,
        is_dir: Optional[bool] = None,
        is_file: Optional[bool] = None,
        is_symlink: Optional[bool] = None,
        stat: Optional[os.stat_result] = None,
    ) -> None:
        self.name = name
        self.path = name if path is None else path
        self._inode = inode
        self._is_dir = is_dir
        self._is_file = is_file
        self._is_symlink = is_symlink
        self._stat = stat

    def inode(self) -> int:
        assert self._inode is not None
        return self._inode

    def is_dir(self, follow_symlinks: bool = True) -> bool:
        assert self._is_dir is not None
        return self._is_dir

    def is_file(self, follow_symlinks: bool = True) -> bool:
        assert self._is_file is not None
        return self._is_file

    def is_symlink(self) -> bool:
        assert self._is_symlink is not None
        return self._is_symlink

    def stat(self, follow_symlinks: bool = True) -> os.stat_result:
        assert self._stat is not None
        return self._stat

    def __str__(self) -> str:
        return f"<_DirEntry {self.name!r}>"

    def __repr__(self) -> str:
        return str(self)

    def __fspath__(self) -> str:
        return self.path


class BaseDirEntry(os.PathLike):
    __slots__ = ("entry",)

    """ You cannot inherit from `os.DirEntry` (type 'nt.DirEntry' is not an acceptable base type),
        so use this wrapper class instead.
    """

    def __init__(self, entry: DirEntryProtocol) -> None:
        self.entry = entry

    @property
    def name(self) -> str:
        return self.entry.name

    @property
    def path(self) -> str:
        return self.entry.path

    def inode(self) -> int:
        return self.entry.inode()

    def is_dir(self, follow_symlinks=True) -> bool:
        return self.entry.is_dir(follow_symlinks=follow_symlinks)

    def is_file(self, follow_symlinks=True) -> bool:
        return self.entry.is_file(follow_symlinks=follow_symlinks)

    def is_symlink(self) -> bool:
        return self.entry.is_symlink()

    def stat(self, follow_symlinks=True) -> os.stat_result:
        return self.entry.stat(follow_symlinks=follow_symlinks)

    def __str__(self) -> str:
        return str(self.entry)

    def __repr__(self) -> str:
        return repr(self.entry)

    def __fspath__(self) -> str:
        return os.fspath(self.entry)


MyDirEntryT = Union[os.DirEntry, BaseDirEntry]


def entrysuffix(entry: MyDirEntryT) -> str:
    return os.path.splitext(entry.name)[1]


def to_dos_device_path(path: str) -> str:
    """Converts a standard absolute Windows path to a DOS device path aka local device paths.
    These paths are not normalized by the Win32 api and can be used to access otherwise unavailable files.
    """

    p = r"(?:(?:\\\\([^\\\/]+))|([a-z]:))([\\\/].*)"
    m = re.match(p, path, re.IGNORECASE)

    if m:
        server, drive, _path = m.groups()
        if drive is not None:
            return "\\\\?\\" + drive + _path
        elif server not in ("?", "."):
            return "\\\\?\\UNC\\" + server + _path

    return path


def to_dos_path(path: str) -> str:
    """Converts a DOS device path aka local device paths to a standard absolute Windows path.
    These paths have better compatibility and familiarity but also more limitations.
    """

    p = r"\\\\(?:\?|\.)\\(?:(unc\\)|([a-z]:))(.*)"
    m = re.match(p, path, re.IGNORECASE)
    if m:
        unc, drive, _path = m.groups()
        if unc:
            return "\\\\" + _path
        else:
            return drive + _path

    return path



from typing import Optional, Tuple, Union

import numpy as np
from sklearn.model_selection import StratifiedShuffleSplit


def train_test_split_indices(
    y: np.ndarray,
    test_size: Optional[Union[int, float]] = None,
    train_size: Optional[Union[int, float]] = None,
    random_state: Optional[int] = None,
) -> Tuple[np.ndarray, np.ndarray]:

    sss = StratifiedShuffleSplit(test_size=test_size, train_size=train_size, random_state=random_state)
    X = np.zeros(len(y))

    train, test = next(sss.split(X, y))
    return train, test
